import re
import sqlparse
from typing import List, Tuple, Iterator, Iterable, Any, Dict

Token = sqlparse.tokens.Token


def _parse_named_placeholder(text: str):
    if text[0] == ':':
        return text[1:], False
    else:
        return text[2:-2], True


class PlaceHolder:
    """
    SQL文中に含まれるIN句を構成する名前付きプレースホルダ情報。
    IN :x
    という記述があったら、「:x」部分の開始位置・長さと認識結果を格納しています。
    :property pos: 開始位置
    :property length: 長さ
    :property name: プレースホルダ名
    :property is_pyformat: %(x)s 形式であったらTrue、:x 形式ならFalse
    :property is_in_paren: IN (:x) の形なら True, IN :x の形ならFalse
    """

    def __init__(self,
                 pos: int,
                 text: str,
                 is_for_in_clause: bool,  # IN句内のプレースホルダであるか
                 is_in_paren: bool  # IN :x ならFalse、IN (:x) ならTrue
                 ):
        self.pos = pos
        self.length = len(text)
        self.name, self.is_pyformat = _parse_named_placeholder(text)
        self.is_for_in_clause = is_for_in_clause
        self.is_in_paren = is_in_paren

    def matches(self, obj):
        """
        指定された名前、またはInVarに一致するプレースホルダであるかを判定します。
        not is_pyformat の場合にはcx_oracleの動作に合わせるため
        大小文字無視して一致判定するようになっています。
        """
        if isinstance(obj, str):
            if self.is_pyformat:
                return self.name == obj
            else:
                return self.name.lower() == obj.lower()
        else:
            if self.is_pyformat:
                return obj.is_pyformat and self.name == obj.name
            else:
                return not obj.is_pyformat and \
                       self.name.lower() == obj.name.lower()

    def find_in_dict(self, d: Dict[str, Any]):
        """
        辞書中に、自分の名前に一致するキーがあればそのキーと値を出力します。
        """
        if self.is_pyformat:
            if self.name in d:
                return self.name, d[self.name]
        else:
            for key in d:
                if self.matches(key):
                    return key, d[key]
        return None

    def generate_replace_string(self, new_names: Iterable[str]):
        """
        このプレースホルダを、指定された名前のプレースホルダの列に変換します。
        この結果をSQLの文字列置換に使用します。
        :param new_names: 名前の列
        """
        if self.is_pyformat:
            return '%(' + ')s,%('.join(new_names) + ')s'
        else:
            return ':' + ',:'.join(new_names)

    def __str__(self):
        return '<%s (%d-%d) %s%s>' % (
            self.name,
            self.pos,
            self.pos + self.length,
            'pyformat' if self.is_pyformat else 'named',
            '' if not self.is_for_in_clause else ' IN()'
            if self.is_in_paren else ' IN'
        )


def extract_placeholders(
        statement: str
) -> List[PlaceHolder]:
    return _parse_expression(
        iter(sqlparse.parse(statement)[0].flatten()))


def _parse_expression(
        token_iterator: Iterator
) -> List[PlaceHolder]:
    """
    括弧に入っていない地の文。
    """
    pos = 0
    placeholders = []  # type: List[PlaceHolder]
    while True:
        try:
            token = next(token_iterator)
        except StopIteration:
            return placeholders
        pos += len(str(token))

        if _is_left_paren(token):
            pos, found_placeholders = \
                _parse_paren(pos, token_iterator)
            placeholders += found_placeholders

        elif _is_keyword_in(token):

            pos, found_placeholders = \
                _parse_in_clause(pos, token_iterator)
            placeholders += found_placeholders

        elif _is_named_placeholder(token):
            placeholders.append(PlaceHolder(pos, str(token),
                                            is_for_in_clause=False,
                                            is_in_paren=False))


def _parse_in_clause_paren(
        pos: int,
        token_iterator: Iterator
) -> Tuple[int, List[PlaceHolder]]:
    """
    INに続く括弧の中。
    """
    placeholders = []  # type: List[PlaceHolder]
    while True:
        try:
            token = next(token_iterator)
        except StopIteration:
            return pos, placeholders
        cur_pos = pos
        pos += len(str(token))

        if _is_right_paren(token):
            return pos, placeholders

        if _is_left_paren(token):
            pos, found_placeholders = \
                _parse_paren(pos, token_iterator)
            placeholders += found_placeholders

        elif _is_named_placeholder(token):

            placeholder = PlaceHolder(cur_pos, str(token),
                                      is_for_in_clause=True,
                                      is_in_paren=True)
            placeholders.append(placeholder)


def _parse_paren(
        pos: int,
        token_iterator: Iterator
) -> Tuple[int, List[PlaceHolder]]:
    """
    括弧の中。
    """
    found = []  # type: List[PlaceHolder]
    while True:
        try:
            token = next(token_iterator)
        except StopIteration:
            return pos, found
        pos += len(str(token))

        if _is_right_paren(token):
            return pos, found

        if _is_left_paren(token):
            pos, found_placeholders = \
                _parse_paren(pos, token_iterator)
            found += found_placeholders

        elif _is_keyword_in(token):

            pos, found_placeholders = \
                _parse_in_clause(pos, token_iterator)
            found += found_placeholders

        elif _is_named_placeholder(token):
            found.append(PlaceHolder(pos, str(token),
                                     is_for_in_clause=False,
                                     is_in_paren=False))


def _parse_in_clause(
        pos: int,
        token_iterator: Iterator
) -> Tuple[int, List[PlaceHolder]]:
    """
    "IN"を呼んだ直後の状態。
    """
    while True:
        try:
            token = next(token_iterator)
        except StopIteration:
            return pos, []
        cur_pos = pos
        pos += len(str(token))

        if _is_left_paren(token):

            return _parse_in_clause_paren(pos, token_iterator)

        elif _is_named_placeholder(token):

            placeholder = PlaceHolder(cur_pos,
                                      str(token),
                                      is_for_in_clause=True,
                                      is_in_paren=False)
            return pos, [placeholder]

        elif not _is_ignorable(token):
            return pos, []


def _is_keyword_in(token):
    return token.ttype == Token.Keyword and str(token).lower() == 'in'


def _is_ignorable(token):
    return token.ttype in (Token.Comment.Multiline,
                           Token.Comment.Single,
                           Token.Text.Whitespace)


def _is_named_placeholder(token):
    s = str(token)
    return token.ttype == Token.Name.Placeholder and \
        ((s[0] == ':' and not re.match(':[0-9]+$', s))
         or s.startswith('%('))


def _is_left_paren(token):
    return token.ttype == Token.Punctuation and str(token) == '('


def _is_right_paren(token):
    return token.ttype == Token.Punctuation and str(token) == ')'



from inspect import currentframe, signature
from typing import List, Tuple, Union, Iterable, Sized, Callable

from sqlinclause.placeholder_extractor import extract_placeholders, PlaceHolder


def expand_sql(
        statement: str,
        params: Union[dict, list, tuple]=None,
        call_depth=1
) -> Tuple[str, Union[dict, list, tuple]]:
    """
    SQL文中でIN句内に名前付きプレースホルダが入っていた場合、実引数リストの数に
    応じて名前付きプレースホルダのカンマ区切り列に展開します。
    
    例： expand_sql_in_clause('select * from some_table where id in :x',
                             {'x': [1,2,3]})
    ⇒
    ('select * from some_table where id in (:e_1,:e_2,:e_3)',
     {'e_1': 1, 'e_2: 2, 'e_3: 3})
    
    :param call_depth: ユーザーコードからの呼び出し深度。直接呼び出す場合は1を指定。
    :param statement: SQL文
    :param params: 実引数辞書。リストやタプルを渡した場合には何も処理せず返します。
    :return: 展開済みのSQL文とそれに適用するための実引数辞書
    """
    if params is None:
        params = {}

    # 辞書型実引数以外には手を触れない
    if not isinstance(params, dict):
        return statement, params

    local_vars = _caller_locals(call_depth)
    placeholders = extract_placeholders(statement)

    supplied_args = dict(params)
    for placeholder in placeholders:
        key_value = placeholder.find_in_dict(params)
        if key_value is None:  # 実引数辞書に必要なキーがなかった！
            local_key_value = placeholder.find_in_dict(local_vars)
            if local_key_value is not None:  # ローカル変数にならあった！
                key, value = local_key_value
                supplied_args[key] = value

    return _expand_sql_in_clause(statement, placeholders, supplied_args)


def expand_sql_in_clause(
        statement: str,
        params: dict
) -> Tuple[str, dict]:
    """
    SQL文中でIN句内に名前付きプレースホルダが入っていた場合、実引数リストの数に
    応じて名前付きプレースホルダのカンマ区切り列に展開します。
    
    例： expand_sql_in_clause('select * from some_table where id in :x',
                             {'x': [1,2,3]})
    ⇒
    ('select * from some_table where id in (:e_1,:e_2,:e_3)',
     {'e_1': 1, 'e_2: 2, 'e_3: 3})
    
    :param statement: SQL文
    :param params: 実引数辞書。
    :return: 展開済みのSQL文とそれに適用するための実引数辞書
    """

    placeholders = extract_placeholders(statement)
    return _expand_sql_in_clause(statement, placeholders, params)


def _expand_sql_in_clause(
        statement: str,
        placeholders: List[PlaceHolder],
        params: dict
) -> Tuple[str, dict]:

    seq = Sequence()
    replacements = []  # type: List[Tuple[PlaceHolder, str]]
    result_args = {}

    for placeholder in placeholders:

        if not placeholder.is_for_in_clause:
            continue

        for replacement in replacements:
            # すでに同じ名前のプレースホルダを置換済みだった場合には、その置換にのっかる
            *_, existing_placeholder, replace_str = replacement
            if existing_placeholder.matches(placeholder):
                replacements.append((placeholder, replace_str))
                break
        else:
            match = placeholder.find_in_dict(params)
            if match is None:
                continue

            key, value = match
            if not isinstance(value, str) and\
                    isinstance(value, Iterable):

                if not isinstance(value, Sized):
                    value = list(value)
                new_names = list(_new_names(placeholders, len(value), seq))
                statement_replace = placeholder.generate_replace_string(new_names)
                result_args.update(zip(new_names, value))
            else:
                statement_replace = placeholder.generate_replace_string([key])

            replacements.append((placeholder, statement_replace))

    result_statement = []
    pos = 0
    for placeholder, text in replacements:
        start = placeholder.pos
        end = placeholder.pos + placeholder.length

        result_statement.append(statement[pos:start])
        if placeholder.is_in_paren:
            result_statement.append(text)
        else:
            result_statement.append('(')
            result_statement.append(text)
            result_statement.append(')')
        pos = end

    result_statement.append(statement[pos:])
    result = ''.join(result_statement)

    for arg_key, arg_value in params.items():
        for placeholder, text in replacements:
            if placeholder.matches(arg_key):
                break
        else:
            result_args[arg_key] = arg_value

    return result, result_args


def _caller_locals(depth: int = 1) -> dict:
    """
    呼び出し元コードの名前空間を辞書として取得します。
    :param depth: _caller_namespace() を呼んでいるコードを0とし、増えるごとに\
    呼び出し元を1つずつ遡っていきます。
    :return:
    """
    f = currentframe().f_back
    for i in range(depth):
        f = f.f_back
    return f.f_locals


class Sequence:
    def __init__(self):
        self.count = 0

    def generate(self):
        self.count += 1
        return self.count


def _new_names(existing: List[PlaceHolder], how_many: int, seq: Sequence):
    for i in range(how_many):
        while True:
            name = 'e_%d' % seq.generate()
            if not any(var.matches(name) for var in existing):
                break
        yield name



#!/usr/bin/env python
# coding: utf-8
from .main import expand_sql, expand_sql_in_clause
from .placeholder_extractor import PlaceHolder, extract_placeholders

__author__  = 'Miura Takuma'
__version__ = '0.0.1'
__license__ = 'MIT'



#!/usr/bin/env python
# coding: utf-8
from setuptools import setup, find_packages
import sqlinclause

setup(
    name             = 'sqlinclause',
    version          = sqlinclause.__version__,
    description      = 'SQL utility to write variable length IN-clause.',
    license          = sqlinclause.__license__,
    author           = sqlinclause.__author__,
    author_email     = 'takuma-miura@m3.com',
    url              = '',
    keywords         = 'SQL IN',
    packages         = ['sqlinclause'],
    install_requires = ['sqlparse'],
)

